<?php 
class permission implements crudable {

	protected static $fields=array(
			'projectid'=>validator::INTEGER,
			'usergroupid'=>validator::INTEGER,
			'type'=>validator::PERMISSIONTYPE
	);
	
	protected static $helpTexts=array(
			'projectid'=>'The ID of the project',
			'usergroupid'=>'The ID of the usergroup',
			'type'=>'What permission is being granted - create, read, update or delete'
	);
	
	public static function getFieldValidations(): array {
		return self::$fields;
	}
	public static function getFieldHelpTexts(): array {
		return self::$helpTexts;
	}

	private static $changePermissions=array('create','update','delete');

    /**
     * @param $id
     * @return array
     * @throws BadRequestException
     * @throws ServerException
     */
	public static function getById($id): array {
		$sqlStatement='SELECT * FROM permission WHERE id=:id';
		return database::queryGetOne($sqlStatement, array(':id'=>$id));
	}

    /**
     * @param $name
     * @throws ServerException
     */
	public static function getByName($name){
		throw new ServerException('getByName not implemented on permission');
	}

    /**
     * @param $key
     * @param $value
     * @param array $request
     * @throws ServerException
     */
	public static function getByProperty($key,$value,$request=array()){
		throw new ServerException('getByProperty not implemented on permission');
	}

    /**
     * @param array $request
     */
	public static function getAll($request=array()): array {
        return array();
	}

    /**
     * @param array $request
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function create($request=array()): array {
		$projectId=$request['projectid'];
		$usergroupId=$request['usergroupid'];
		$type=$request['type'];
		$project=project::getbyid($projectId); //throws exception if the project can't be read
		usergroup::getbyid($usergroupId); //throws exception if the usergroup can't be read
		validator::validate($type, $request['type'], self::$fields['type']);
		if(!session::isAdmin() && session::getUserId()!=$project['owner']){
			throw new ForbiddenException('Only administrators and project owners can change groups\' project permissions');
		}
		$allPermissions=['read','create','update','delete'];
		if(!in_array($type, $allPermissions)){
		    throw new BadRequestException('Bad permission type ('.$type.') in permission::create');
        }
		//Delete all the old permissions for this project/group combination
		foreach($allPermissions as $oldPermission){
            $params=array(':projectid'=>$projectId, ':usergroupid'=>$usergroupId, ':type'=>$oldPermission);
            $sqlStatement='DELETE FROM permission WHERE projectid=:projectid AND usergroupid=:usergroupid AND type=:type';
            database::query($sqlStatement, $params);
        }
		//If read, create only read permission, else create read,create,update,delete permissions
		$created=[];
		if('read'==$type){
		    $newPermissions=['read'];
        } else {
		    $newPermissions=$allPermissions;
        }
		foreach ($newPermissions as $newPermission){
            $sqlStatement='INSERT INTO permission(projectid,usergroupid,type) VALUES(:projectid,:usergroupid,:type)';
            $params=array(':projectid'=>$projectId, ':usergroupid'=>$usergroupId, ':type'=>$newPermission);
            database::query($sqlStatement, $params);
            $lastId=database::getLastInsertId();
            $created[]=permission::getById($lastId);
        }
		return array('created'=>$created);
	}

    /**
     * @param $id
     * @param array $request
     * @throws ServerException
     */
	public static function update($id, $request=array()){
		throw new ServerException('Update not implemented on permission');
	}

    /**
     * Deletes the permission. If permission type is one of "create","update","delete",
     * all create/update/delete permissions for this project and usergroup will be deleted.
     * @param $id
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
	public static function delete($id): array {
		$permission=self::getbyid($id);
		$project=project::getbyid($permission['projectid']);
		$usergroup=usergroup::getbyid($permission['usergroupid']);
		$type=$permission['type'];
		if(empty($permission) || empty($project) || empty($usergroup) ){ 
			throw new NotFoundException('Permission does not exist or you cannot read it'); 
		}
		if(!session::isAdmin() && session::getUserId()!=$project['owner']){
			throw new ForbiddenException('Only administrators and project owners can change groups\' project permissions');
		}

		$deletedPermissions=[];
		$permissionsToDelete=static::$changePermissions;
		if('read'==$type){
		    $permissionsToDelete[]='read';
        } else if(!in_array($type, static::$changePermissions)) {
            throw new ServerException('Bad permission type ('.$type.') in permission::delete');
        }
		foreach ($permissionsToDelete as $permissionToDelete) {
            $existing=database::queryGetOne('SELECT * FROM permission 
                WHERE projectid=:projectid AND usergroupid=:usergroupid AND type=:type',
                array(
                    ':projectid'=>$permission['projectid'],
                    ':usergroupid'=>$permission['usergroupid'],
                    ':type'=>$permissionToDelete
                )
            );
            if(!empty($existing)){
                $deletedPermissions[]=$existing;
                $sqlStatement='DELETE FROM permission WHERE id=:id';
                database::query($sqlStatement, array(':id'=>$existing['id']));
            }
        }
        return array('deleted'=>$deletedPermissions);
	}

	public static function canCreate(): bool {
		return true; // and let the actual create sort it out
	}

    /**
     * @param $id
     * @return bool
     */
	public static function canUpdate($id): bool {
		return true; // and let the actual update sort it out
	}
	
}